#!/usr/bin/env python

import os
import sys
import time
import types
import re
import logging
import copy
from urllib import unquote as unescapeURL

from xpcom import components, COMException
from xpcom.server import UnwrapObject
from xpcom._xpcom import (PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC,
                          getProxyForObject)

import timeline

# Pref names each checker should have, will be monitored when the checker
# is first added and gets automatically updated through a pref observer.
monitoredPrefNames = { "enabledPrefName": types.BooleanType,
                       "executablePrefName": types.StringType,
                       "backgroundEnabledPrefName": types.BooleanType,
                       "backgroundDurationPrefName": types.LongType,
                       "recursivePrefName": types.BooleanType }

class KoFileCheckerBase:

    _com_interfaces_ = [components.interfaces.nsIObserver,
                        components.interfaces.koIFileStatusChecker]

    # Save have to look this up all over the status checker code.
    _is_windows = sys.platform.startswith("win")

    def __init__(self, type, name):
        self.type = type
        self.name = name

        # Dictionary of when a URI was last checked.
        # Note: On Windows, the URI must be lowered, because it's a case
        #       insensitive filesystem and we never can be sure which case
        #       styling will be used. You should use the "_norm_uri_cache_key"
        #       method for this (below).
        #       http://bugs.activestate.com/show_bug.cgi?id=74339
        self._lastChecked = {}

        # How prefName and matching attribute values work:
        #  When a checker is initialized, the prefName(s) are checked and the
        #  pref value is stored into the correctponding attribute name. A pref
        #  observer is added on the preference and when/if this preference ever
        #  changes, then the checker's value will be automatically updated.
        # Example for CVS:
        #    backgroundDurationPrefName = "cvsBackgroundMinutes"
        #    backgroundDuration = 10

        self.enabled = True
        self.enabledPrefName = None
        self.backgroundEnabled = False
        self.backgroundEnabledPrefName = None
        # We store the duration in seconds.
        self.backgroundDuration = 15 * 60   # 15 minutes
        self.backgroundDurationPrefName = None
        self.recursive = 0
        self.recursivePrefName = None
        self.executable = None
        self.executablePrefName = None

        self.log = logging.getLogger('Ko%sChecker::%s' % (self.type.capitalize(), self.name))
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs

        # Copy across names from the interface in to the local instance.
        self.REASON_BACKGROUND_CHECK = components.interfaces.koIFileStatusChecker.REASON_BACKGROUND_CHECK
        self.REASON_ONFOCUS_CHECK = components.interfaces.koIFileStatusChecker.REASON_ONFOCUS_CHECK
        self.REASON_FILE_CHANGED = components.interfaces.koIFileStatusChecker.REASON_FILE_CHANGED
        self.REASON_FORCED_CHECK = components.interfaces.koIFileStatusChecker.REASON_FORCED_CHECK

    ##
    # Helper function to ensure the cache key "uri" is consistently the same,
    # no matter how the platform handles filename case sensitivity or
    # whether the uri has been escaped.
    def _norm_uri_cache_key(self, uri):
        uri = unescapeURL(uri)
        if self._is_windows:
            return uri.lower()
        return uri

    ##
    # Helper function to check if the nsIURI object has a UNC path.
    def _is_nsURI_UNC(self, nsUri):
        return (self._is_windows and nsUri.scheme == 'file' and nsUri.host)

    ##
    # Helper function to ensure the cache is completely cleared.
    def _invalidateAllCaches(self):
        self._lastChecked = {}

    #  Interface method
    def initialize(self):
        prefObserverSvc = self._globalPrefs.prefObserverService
        for prefSetting, prefType in monitoredPrefNames.items():
            prefName = getattr(self, prefSetting)
            if prefName:
                variableName = prefSetting.replace("PrefName", "")
                # Update from the preference
                if prefType == types.BooleanType:
                    setattr(self, variableName,
                            self._globalPrefs.getBooleanPref(prefName))
                elif prefType == types.SliceType:
                    setattr(self, variableName,
                            self._globalPrefs.getStringPref(prefName))
                elif prefType == types.LongType:
                    value = self._globalPrefs.getLongPref(prefName)
                    if variableName == "backgroundDuration":
                        # Convert from minutes to seconds.
                        value *= 60
                    setattr(self, variableName, value)

                # Listen for pref changes
                prefObserverSvc.addObserver(self, prefName, 0)

    ##
    # nsIObserver interface: listens for preference changes
    # @private
    def observe(self, subject, topic, data):
        self.log.debug("observing event %s:%s" % (topic, data))
        if not topic:
            return

        # data is actually the pref name that was changed.
        if topic == self.executablePrefName:
            executable = self._globalPrefs.getStringPref(topic)
            if executable != self.executable:
                self.setExecutable(executable)
                self._invalidateAllCaches()
                fileStatusSvc = components.classes["@activestate.com/koFileStatusService;1"].\
                                    getService(components.interfaces.koIFileStatusService)
                fileStatusSvc.updateStatusForAllFiles(self.REASON_ONFOCUS_CHECK)
        elif topic == self.enabledPrefName:
            enabled = self._globalPrefs.getBooleanPref(self.enabledPrefName)
            if enabled != self.enabled:
                self.enabled = enabled
                self._invalidateAllCaches()
                fileStatusSvc = components.classes["@activestate.com/koFileStatusService;1"].\
                                    getService(components.interfaces.koIFileStatusService)
                fileStatusSvc.updateStatusForAllFiles(self.REASON_ONFOCUS_CHECK)
        elif topic == self.backgroundEnabledPrefName:
            backgroundEnabled = self._globalPrefs.getBooleanPref(topic)
            if backgroundEnabled != self.backgroundEnabled:
                self.backgroundEnabled = backgroundEnabled
        elif topic == self.backgroundDurationPrefName:
            backgroundDuration = self._globalPrefs.getBooleanPref(topic) * 60
            if backgroundDuration != self.backgroundDuration:
                self.backgroundDuration = backgroundDuration
        elif topic == self.recursivePrefName:
            self.recursive =  self._globalPrefs.getBooleanPref(topic)

    #  Interface method
    def shutdown(self):
        # Remove pref listeners
        prefObserverSvc = self._globalPrefs.prefObserverService
        for prefSetting in monitoredPrefNames:
            prefName = getattr(self, prefSetting)
            if prefName:
                try:
                    prefObserverSvc.removeObserver(self, prefName, 0)
                except:
                    # prefs shutdown already?
                    self.log.debug("Unable to remove prefs observers")

    #  Interface method
    def isActive(self):
        return self.enabled

    #  Interface method
    def isBackgroundCheckingEnabled(self):
        return False

    #  Interface method
    def needsToReCheckFileStatus(self, koIFile, reason):
        return True

    #  Interface method
    def updateFileStatus(self, koIFile, reason):
        return None

    def setExecutable(self, executable):
        self.executable = executable

class KoDiskFileChecker(KoFileCheckerBase):

    _reg_clsid_ = "{4871ae1f-edb2-4e2b-b35f-85109aea68ef}"
    _reg_contractid_ = "@activestate.com/koFileStatusChecker?type=disk;1"
    _reg_desc_ = "Komodo Disk Status Checker"
    #_reg_categories_ = [
    #     ("category-komodo-file-status",      "disk"),
    #     ]

    ranking_weight = 5

    def __init__(self):
        timeline.enter('KoDiskFileChecker.__init__')
        KoFileCheckerBase.__init__(self, 'disk', 'Disk')
        self.enabledPrefName = 'diskStatusEnabled'
        self.backgroundEnabledPrefName = 'diskBackgroundCheck'
        self.backgroundDurationPrefName = 'diskBackgroundMinutes'
        timeline.leave('KoDiskFileChecker.__init__')

    def isBackgroundCheckingEnabled(self):
        return True

    def updateFileStatus(self, koIFile, reason):
        if koIFile.isLocal:
            time_now = time.time()
            cache_key = self._norm_uri_cache_key(koIFile.URI)
            if reason == self.REASON_BACKGROUND_CHECK and \
               (self._lastChecked.get(cache_key, 0) >=
                (time_now - self.backgroundDuration)):
                return 0
            self._lastChecked[cache_key] = time_now
            return koIFile.hasChanged
        elif reason == self.REASON_FORCED_CHECK:
            # Forced, hasChanged will cause a refresh from the remote server.
            return koIFile.hasChanged
        return 0


class KoSCCChecker(KoFileCheckerBase):

    # This is the dictionary structure for "koIFileEx.scc", it should be kept
    # in sync with what URIlib::set_scc and URIlib::clearscc do.
    baseFileSCCInfo = {
        'sccType':'',
        'sccDirType': '',
        'sccExclude':0,
        'sccHaveOnDisk':0,
        'sccLocalRevision':'',
        'sccRevDate':'',
        'sccDepotRevision':'',
        'sccNeedSync':0,
        'sccSync':0,
        'sccOk':0,
        'sccConflict':0,
        'sccAction':'',
        'sccStatus':'',
        'sccChange':''
    }

    def __init__(self, name='unknown'):
        timeline.enter('KoSCCChecker.__init__')
        KoFileCheckerBase.__init__(self, 'scc', name)
        #self.log.setLevel(logging.DEBUG)

        # Dictionary of cached scc information for a given URI.
        # Note: On Windows, the URI must be lowered, because it's a case
        #       insensitive filesystem and we never can be sure which case
        #       styling will be used.
        #       http://bugs.activestate.com/show_bug.cgi?id=74339
        self._cached_info = {}

        self.type = 'scc'
        # SCC name, one of "p4", "cvs" or "svn"
        self.name = name
        # Error details.
        self.error = None
        self.error_detail = None
        self.last_error_msg = None
        self.last_error_detail = None
        # Special marker for remote scc.
        self._sccRemote = 0
        # Prefs used for checking customizable settings
        self.executablePrefName = ''
        self.enabledPrefName = ''
        self.backgroundEnabledPrefName = ''
        self.backgroundDurationPrefName = ''
        self.recursivePrefName = ''
        self.svc = components.classes["@activestate.com/koSCC?type=%s;1" % (name)].\
            getService(components.interfaces.koISCC)
        timeline.leave('KoSCCChecker.__init__')
        self.iosvc = components.classes["@mozilla.org/network/io-service;1"].\
            getService(components.interfaces.nsIIOService)
        # Global observer service.
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self._observerProxy = getProxyForObject(None,
            components.interfaces.nsIObserverService, self._observerSvc,
            PROXY_ALWAYS | PROXY_SYNC)

    def initialize(self):
        KoFileCheckerBase.initialize(self)
        # Executable is better coming from the service, as it knows better.
        self.setExecutable(self.svc.executable)

    # Plan A: In-use
    # Store all retrieved scc information under the file/dir path in the
    # _cached_info dictionary. Update the _lastChecked dict when it's updated.
    # The _cache_info objects will hold the information necessary for koIFile.
    # To get the last updated information for a scc file just requires looking
    # at _lastChecked with the file path.
    #
    # Plan B: Thinking about it
    # Given a file URI "uri", we obtain the directory for this file "dirURI".
    # Store all retrieved scc file information for "dirURI" in dictionary
    # "dict". This dictionary "dict" will be added into the fileStatusChecker
    # "_lastChecked" dictionary using the key "dirURI". To get the last updated
    # information for a file "uri" requires getting "dict" from the
    # "_lastChecked" using the "dirURI" key. The scc info is stored in "dict"
    # with the "uri" key. The invalidation of "uri" is then much easier, simply
    # drop whole the "dirURI" dictionary out of "_lastChecked" and then re-run
    # the scc update on "dirURI".
    #

    ##
    # Helper function to ensure the cache is completely cleared.
    def _invalidateAllCaches(self):
        KoFileCheckerBase._invalidateAllCaches(self)
        self._cached_info = {}

    # Same as base class
    #def setExecutable(self, executable):
    #    self.executable = executable

    ##
    # Notify an error has occured. The error and detail will be displayed in
    # the "SCC Output" tab.
    # @private
    # @param error {string} The error that occured.
    # @param detail {string} Additional information about the error.
    def notifyError(self, error, detail=''):
        self.log.debug(detail)
        # Ignore if this error is the same as the last one shown.
        if self.last_error_msg == error and self.last_error_detail == detail:
            return
        self.error = components.classes["@activestate.com/koStatusMessage;1"]\
                .createInstance(components.interfaces.koIStatusMessage)
        self.error.category = "scc_error_message"
        self.error.msg = error
        self.error.timeout = 5000
        self.error.highlight = 1
        self.error_detail = str(detail)
        # XXX 'scc_status_message' is evil (see bug 20568).
        #print "attempting to send notification"

        try:
            self._observerProxy.notifyObservers(self.error,
                                                'scc_status_message',
                                                str(detail))
        except COMException, e:
            pass

    def clearErrors(self):
        self.error = None
        self.error_detail = None

    # Overriding parent isActive
    def isActive(self):
        return self.enabled and self.svc.isFunctional

    # Overriding parent isBackgroundCheckingEnabled
    def isBackgroundCheckingEnabled(self):
        return True

    def isSupportedFile(self, koIFile):
        return koIFile.isLocal and (koIFile.isFile or koIFile.isDirectory)

    # Remove any known SCC information for this path if:
    #  1) it's time to re-check this because of age
    #  2) the provided reason indicates we need fresh information
    # Returns true if the path was invalidated and thus should be checked
    #   koIFile - unwrapped koIFile XPCOM object
    # Note: On Windows, all path names must be lowered, because it's a case
    #       insensitive filesystem and we never can be sure which case styling
    #       will be used.
    def needsToReCheckFileStatus(self, koIFile, reason):
        # XXX - Need better test for this
        #if isinstance(koIFile, components.interfaces.koIFile):
        koIFile = UnwrapObject(koIFile)

        if not self.isSupportedFile(koIFile):
            return 0

        # If this file is already marked with a scc type,
        # we only need to re-check with the same checker.
        if ((koIFile.sccType and self.name != koIFile.sccType) or
            (koIFile.sccDirType and self.name != koIFile.sccDirType)):
            self.log.debug("Checker already set to another")
            return 0

        uri_cache_key = self._norm_uri_cache_key(koIFile.URI)
        self.log.debug("%s needsToReCheckFileStatus (reason=%r) uri: %r",
                       self.name, reason, uri_cache_key)

        if reason in (self.REASON_ONFOCUS_CHECK,
                      self.REASON_FORCED_CHECK,
                      self.REASON_FILE_CHANGED):
            force = True
        else:
            force = False

        diruri_cache_key = uri_cache_key
        if koIFile.isFile:
            diruri_cache_key = os.path.dirname(uri_cache_key)
        if not diruri_cache_key.endswith("/"):
            # Ensure it's compatible with nsURI.
            diruri_cache_key += "/"
        dirLastChecked = self._lastChecked.get(diruri_cache_key)
        if dirLastChecked is not None:
            previousCheck = time.time() - self.backgroundDuration
            if force or dirLastChecked < previousCheck:
                self.log.debug("%s: removing cached info", self.name)
                # Remove the cache information, it's no longer valid
                self._lastChecked.pop(diruri_cache_key, None)
                self._cached_info.pop(uri_cache_key, None)
                return 1
            elif self._cached_info.get(uri_cache_key) != koIFile.scc:
                # This url has never been checked, or it's cache has
                # expired, or it's scc status has changed.
                return 1
            self.log.debug("%s: cached has not yet expired", self.name)
            return 0
        else:
            self.log.debug("%s: no cached info found", self.name)
        return 1

    # Check the SCC status for the given file
    # Returns true if the path was under scc and was successfully updated
    #   koIFile   - unwrapped koIFile XPCOM object
    #   reason    - why the check is being made on this file.
    # Note: SCC subclasses must implement updateSCCInfo() in order
    #       to update their known scc information.
    def updateFileStatus(self, koIFile, reason):
        # XXX - Need better test for this
        #if isinstance(koIFile, components.interfaces.koIFile):
        koIFile = UnwrapObject(koIFile)

        oldscc = copy.deepcopy(koIFile.scc)
        if not self.isActive():
            # Empty the koISCC fields and return.
            koIFile.scc = {}
            return koIFile.scc != oldscc

        uri = koIFile.URI
        uri_cache_key = self._norm_uri_cache_key(uri)
        nsUri = self.iosvc.newURI(uri, None, None)
        self.log.debug("%s: Checking %r", self.name, uri)

        # If it's a file, get the parent directory, we'll use that mostly
        dir_nsUri = nsUri
        dir_uri = uri
        if not koIFile.isDirectory:
            dir_nsUri = self.iosvc.newURI("./",
                                          nsUri.originCharset,
                                          nsUri)
            dir_uri = dir_nsUri.spec

        # If we've already checked this path, then we don't need to do
        # it again
        diruri_cache_key = self._norm_uri_cache_key(dir_uri)
        isUpToDate = self._lastChecked.get(diruri_cache_key, 0)
        if isUpToDate:
            if uri_cache_key not in self._cached_info:
                self.log.debug("%s: Already checked uri: not under SCC",
                               self.name)
                return 0
            # Else, we know this directory "dir_uri" is under SCC, set the
            # file scc information from the cache then.
            koSccInfo = self._cached_info.get(uri_cache_key)
            if not koSccInfo:
                # Will reset scc and then update this one attribute.
                koIFile.scc = {'sccDirType': self.name}
                self.log.debug("%s: File not under SCC, but can be added.",
                               self.name)
                self._cached_info[uri_cache_key] = koIFile.scc
                return koIFile.scc != oldscc

            sccInfoHasChanged = (oldscc != koSccInfo)
            if sccInfoHasChanged:
                koIFile.scc = koSccInfo
            #self._updateFileFromCachedInfo(koIFile, _sccInfo)
            self.log.debug("%s: File is under SCC (cached), has it changed: %r",
                           self.name, sccInfoHasChanged)

            return sccInfoHasChanged

        # We're checking this path now, so update the checked time
        self._lastChecked[diruri_cache_key] = time.time()
        koSccInfo = None

        self.log.debug("%s: No cached info, running scc update now", self.name)
        if not self.updateSCCInfo(dir_nsUri, reason):
            # Failed to update SCC information
            self.log.debug("%s: No scc information found", self.name)
            return 0

        # This ensures that the directory path is marked as a SCC folder
        self._cached_info[diruri_cache_key] = {'sccDirType': self.name}

        koSccInfo = self._cached_info.get(uri_cache_key)
        if not koSccInfo:
            # It's in a SCC directory, but it's not added yet
            # We record the dirType so that it can be added later.
            # Will reset scc and then update this one attribute.
            koIFile.scc = {'sccDirType': self.name}
            self.log.debug("%s: File not under SCC, but can be added.",
                           self.name)
            self._cached_info[uri_cache_key] = koIFile.scc
            return oldscc != koIFile.scc

        # Updated successfully, we place in the updated scc info now
        sccInfoHasChanged = (oldscc != koSccInfo)
        if sccInfoHasChanged:
            koIFile.scc = koSccInfo
        self.log.debug("%s: Is under SCC, has it changed: %r",
                      self.name, sccInfoHasChanged)
        return sccInfoHasChanged

    # Update the known SCC information for the given path.
    # Virtual method, must be implemented by the inheriting SCC class.
    def updateSCCInfo(self, dir_nsUri, reason):
        raise NotImplementedError("Function must be implemented by a subclass")

    def printCache(self):
        print "\n%r cache:" % (self.name, )
        for name, data in self._cached_info.items():
            print "  name: %r" % (name, )
            print data
            print
        print

