/* Copyright (c) 2003-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* Komodo's Debug Session tab handling code.
 *
 * Here is the deal: When you start a debug session in Komodo you get a new
 * Debug Session Tab in the bottom pane. This is it here. The Komodo code that
 * starts a debug session calls into here to get a new tab. A handler XPCOM
 * component (DebugSessionTabManager) is created for the tab and Komodo's
 * debugging works with the debug tab through it's API.
 *
 * Usage:
 *      var tabMgr = this.manager.getTabMgr();
 *      if (!tabMgr) {
 *          // Failed to create a tab and manager. The user was notified via
 *          // an alert. The debug session should be aborted.
 *      }
 *      tabMgr.configure(session);
 *      tabMgr.sessionInit();
 *      tabMgr.sessionStart();
 *      tabMgr.sessionEnd();
 *      ...use the koIDebugSessionTabManager interface...
 *
 */

//---- globals


if (typeof(ko) == 'undefined') {
    var ko = {};
}
if (typeof(ko.dbg) == 'undefined') {
    ko.dbg = {};
}

(function() { /* ko.dbg */

const MAX_DEBUG_SESSION_TABS = 100;
var log = ko.logging.getLogger("debugSessionTab");
//log.setLevel(ko.logging.LOG_DEBUG);
const NS_XUL = 'http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul';

// Cache of tab managers.
// - Maps a tab number/id to a DebugSessionTabManager() instance.
// - If managers[<tab-number>] does not exist then no such tab/tabpanel has
//   been created.
// - If managers[<tab-number>]==null then there is a tab/tabpanel left over
//   from a previous debug session, it is just collapsed.
// - If managers[<tab-number>]==<DebugSessionTabManager-instance> then
//   there is an existing debug session in that tab/tabpanel.
//   XXX Note that this debug session may no longer be active in which
//       case we *may* want to re-use this tab. Hmmm. How about adding a
//       "pin" icon to an inactive tab to ensure it doesn't get re-used.
//       Could do the same for the Find Results windows.


//---- the DebugSessionTab component

function DebugSessionTabManager(id) {
    this._outputTmpFile = null;
    this.initialize(id);
}
DebugSessionTabManager.prototype.constructor = DebugSessionTabManager;

DebugSessionTabManager.prototype.QueryInterface = function (iid) {
    if (!iid.equals(Components.interfaces.koIDebugSessionTabManager) &&
        !iid.equals(Components.interfaces.nsISupports)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
}


DebugSessionTabManager.prototype.initialize = function(id)
{
    log.debug("DebugSessionTabManager["+id+"].initialize()");
    try {
        // Number identifying this DebugSessionTabManager from others.
        this.id = id;
        this._idprefix = "debugsession"+this.id;

        this.name = null;
        this.contextNames = null;
        // Cache of useful widget references in this tab.
        this.tab = document.getElementById(this._idprefix+"-tab");
        this.tabPanel = document.getElementById(this._idprefix+"-tabpanel");
        this.tabPanel.onLoad(this, true);

        this._watched = {};
        this._varState = {};
        this.currentStackFrameIndex = -1;
        this.resetTabName();
        this.clear();
        this.isInteractive = false;
        this.basedir = null;
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.initVarTabs = function(/* XXX arguments here? */)
{
    log.debug("DebugSessionTabManager["+this.id+"].initVarTabs(...)");
    try {
        this.tabPanel.watchVariablesTree.onLoad(new variableTreeView(this, 'Watched'));

        // get the context names now
        if (!this.contextNames) {
            var tmp = new Object();
            this.session.dbg.contextNames(tmp, new Object());
            this.contextNames = tmp.value;
        }

        // XXX for debugging only
        log.debug("  number of contexts: "+this.contextNames.length);
        if (this.contextNames.length < 0) return;

        // set up the trees
        // we do not include the watched vars tab here, as this stuff
        // needs to be dynamic
        this._varTabs = {};
        for (var i = 0; i < this.contextNames.length; i++) {
            log.debug("  context: "+this.contextNames[i].name);
            var contextTab = this.tabPanel.addVariablePanel(this.contextNames[i])
            contextTab[1].onLoad(new variableTreeView(this, this.contextNames[i].name));
            this._varTabs[this.contextNames[i].name] = contextTab;
        }

    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.configure = function(session)
{
    log.debug("DebugSessionTabManager["+this.id+"].configure(...)");
    try {
        if (typeof(session) == 'undefined')
            session = null;
        this.session = session;
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.show = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].show()");
    try {
        // If necessary, open output pane in komodo.xul
        ko.uilayout.ensureOutputPaneShown();

        // switch to proper tab
        var tabsWidget = document.getElementById("output_tabs");
        var tabWidget = document.getElementById(this._idprefix+"-tab");
        tabsWidget.selectedItem = tabWidget;
    } catch(ex) {
        log.exception(ex);
    }
}


DebugSessionTabManager.prototype.clear = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].clear()");
    try {
        this.tabPanel.terminalView.clear();
    } catch(ex) {
        log.exception(ex);
    }
}


DebugSessionTabManager.prototype.canClose = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].canClose()");
    try {
        // Warn if this tab is still active, offer to stop it if so.
        if (!this.session) return true;

        var sessionType = 'debugging session';
        var doNotAskPref = 'debugger_confirm_close';
        if (this.session.isInteractive) {
            sessionType = 'interactive shell';
            doNotAskPref = 'interactiveshell_confirm_close';
        }

        var answer = ko.dialogs.customButtons(
            "This "+sessionType+" is still in progress. The session "+
            "must be stopped before the tab can be closed.",
            ["&Kill Session and Close Tab", "Cancel"],
            "Cancel",  // default button
            null, // text
            null, // title
            doNotAskPref);
        if (answer == "Kill Session and Close Tab") {
            return true;
        } else { // answer == "Cancel"
            return false;
        }
    } catch(ex) {
        log.exception(ex);
    }
    return false;
}

DebugSessionTabManager.prototype.forceClose = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].forceClose()");
    try {
        this.tabPanel.releaseTerminal();
        if (!this.session) return;
        this.session.terminate();
    } catch(e) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.resetTabName = function()
{
    var name = 'Output';
    if (this.session) {
        if (this.session.invoke_type == ko.dbg.invocation.INVOKE_INTERACTIVE)
            name = this.session.dbg.languageName+' Shell';
        else if (this.session.invoke_type == ko.dbg.invocation.INVOKE_RUN)
            name = "Run Output";
        else if (this.session.invoke_type == ko.dbg.invocation.INVOKE_DEBUG)
            name = "Debug Output";
    }
    this.tab.setAttribute('label', name);
}

/* sessionInit starts prior to having a debugger connection,
   sessionStart occurs after a having a debugger connection.
   This is important becuase prior to connection we need to have
   some things in place, such as the terminalHandler calling
   startSession to configure io.  At this point the session.dbg
   variable is NOT available.  Since some items may need this for
   initialization (such as the variable viewers), we now have
   these two functions.  */
DebugSessionTabManager.prototype.sessionInit = function(isInteractive)
{
    log.debug("DebugSessionTabManager["+this.id+"].sessionInit()");
    try {
        this.tabPanel.terminalView.clear();
        this.tabPanel.terminalView.startSession();
        this.isInteractive = isInteractive;
        this.tabPanel.setInteractive(isInteractive);
        this.name = null;
        this.resetTabName();
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.sessionStart = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].startSession()");
    try {
        // startup items that require access to session.dbg
        this.initVarTabs();
        this.resetTabName();
        this.loadHTMLView();
        this.basedir = null;
        if (this.session && this.session.filename) {
            var osPathSvc = Components.classes["@activestate.com/koOsPath;1"].
                  getService(Components.interfaces.koIOsPath);
            this.basedir = osPathSvc.dirname(this.session.filename);
        }
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.sessionEnd = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].endSession()");
    try {
        this.tabPanel.terminalView.endSession();
        this.removeHTMLTempFile();
        this.clearThreadMenu();
        this.clearCallStackList();
        this.clearVariables();
        this.tabPanel.removeVariablePanels();
        this.session = null;
        ko.dbg.tabManager.releaseTabMgr(this);
        this.currentStackFrameIndex = -1;
        this.updateWatchedVars();
        if (this.isInteractive) {
            if (gPrefs.getBooleanPref("interactiveShellCloseOnEnd")) {
                // Close interactive session tabs when done
                ko.dbg.tabManager.closeCurrent();
                // focus in the editor.
                if (ko.views.manager.currentView) {
                    ko.views.manager.currentView.setFocus();
                }
            }
        }
        this.isInteractive = false;
        this.contextNames = null;
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.clearVariables = function() {
    log.debug("DebugSessionTabManager["+this.id+"].clearVariables()");
    try {
        for (var context in this._varTabs) {
            this._varTabs[context][1].view.clearList();
        }
    } catch(ex) {
        log.exception(ex);
    }
    delete this._varTabs;
}

DebugSessionTabManager.prototype.updateVariables = function() {
    log.debug("DebugSessionTabManager["+this.id+"].updateVariables()");
    try {
        if (!this.session) return null;
        var state = this.session.getState();
        if (state != ko.dbg.manager.STATE_BREAK && state != ko.dbg.manager.STATE_INTERACTIVE) return null;
        // update information displayed about variables defined in the
        // current stack frame

        // only update variables in the currently visible tab
        var tab = this.tabPanel.variablesTabs.selectedItem;
        var currentContext = tab.getAttribute('label');

        if (!(currentContext in this._varTabs))
            return this.updateWatchedVars()
        var varTab = this._varTabs[currentContext][1];
        // get the current selection so we can reset that
        // after updating the variable list
        var ci = varTab.view.mSelection.currentIndex;
        varTab.view.clearList();
        
        // get the current context names
        if (!this.contextNames) {
            var tmp = new Object();
            this.session.dbg.contextNames(tmp, new Object());
            this.contextNames = tmp.value;
        }
        var context;
        for (var i = 0; i < this.contextNames.length; i++) {
            context = this.contextNames[i];

            // we'll only update the visible context
            if (currentContext != context.name) continue;

            tmp = new Object();
            this.session.dbg.contextGet(context.id,
                                        this.currentStackFrameIndex,
                                        tmp, new Object());
            var debugProps = tmp.value;
            varTab.view.buildList(debugProps, this.currentStackFrameIndex);
            break;
        }
        varTab.view.selectIndex(ci);
    } catch(ex) {
        log.exception(ex);
    }
    return null;
}

DebugSessionTabManager.prototype.updateWatchedVars = function() {
    log.debug("DebugSessionTabManager["+this.id+"].updateWatchedVars()");
    try {
        if (!this.tabPanel.watchVariablesTree.view) return;
        this.tabPanel.watchVariablesTree.view.clearList();
        // clear the expression variables
        var vars = [];
        var expression;
        // get them again
        if (this.session) {
            for (expression in this._watched) {
                log.debug('+++ showing expression ['+expression+']\n');
                var p = this.session.dbg.propertyGetEx(
                            0, this.currentStackFrameIndex,
                            expression, 0, 0, 0, "");
                this._watched[expression] = p;
                vars.push(p);
            }
        } else {
            for (expression in this._watched) {
                log.debug('--- showing expression ['+expression+']\n');
                this._watched[expression] = null;
                vars.push(expression);
            }
        }
        // display them
        this.tabPanel.watchVariablesTree.view.buildList(vars);
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.addWatchedVar = function(expression, value) {
    log.debug("DebugSessionTabManager["+this.id+"].addWatchedVar("+expression+")");
    try {
        this._watched[expression] = null;
        if (this.session) {
            if (typeof(value) != 'undefined') {
                this.session.dbg.propertySet(expression, value);
            }
        }
        this.updateWatchedVars();
    } catch(ex) {
        log.exception(ex);
    }
}


DebugSessionTabManager.prototype.removeWatchedVarName = function(expression)
{
    log.debug("DebugSessionTabManager["+this.id+"].removeWatchedVarName("+expression+")");
    try {
        if (typeof(this._watched[expression]) != 'undefined') {
            delete this._watched[expression];
            return true;
        }
    } catch(ex) {
        log.exception(ex);
    }
    return false;
}

DebugSessionTabManager.prototype.haveWatchedVar = function(expression)
{
    log.debug("DebugSessionTabManager["+this.id+"].haveWatchedVar("+expression+")");
    try {
        var haveit = (typeof(this._watched[expression]) != 'undefined')
    } catch(ex) {
        log.exception(ex);
    }
    return haveit;
}

/**
 * take the text from the debugger output scintilla, put it into the
 * iframe, and switch
 */
DebugSessionTabManager.prototype.removeHTMLTempFile = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].removeHTMLTempFile()");
    try {
        if (this._outputTmpName) {
            var tmpFileSvc = Components.classes["@activestate.com/koFileService;1"]
                             .getService(Components.interfaces.koIFileService)
            tmpFileSvc.deleteTempFile(this._outputTmpName, true);
            this._outputTmpName = null;
        }
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.loadHTMLView = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].renderOutputAsHTML()");
    try {
        this.removeHTMLTempFile();
        var doc;
        var osPathSvc = Components.classes["@activestate.com/koOsPath;1"].
                  getService(Components.interfaces.koIOsPath);
        var docSvc = Components.classes['@activestate.com/koDocumentService;1']
                    .getService(Components.interfaces.koIDocumentService);

        // get the text from scintilla, stuff it into the iframe, and switch the deck
        var text = this.tabPanel.terminalView.scimoz.text;
        if (text) {
            var tmpFileSvc = Components.classes["@activestate.com/koFileService;1"]
                             .getService(Components.interfaces.koIFileService)

            if (this.basedir) {
                this._outputTmpName = tmpFileSvc.makeTempNameInDir(this.basedir,".html");
            } else {
                this._outputTmpName = tmpFileSvc.makeTempName(".html");
            }

            doc = docSvc.createDocumentFromURI(this._outputTmpName);
            doc.setBufferAndEncoding(text, this.tabPanel.terminalView.scintilla.encoding);
            doc.save(true);
            this.tabPanel.browser.initWithDocument(doc, false);
        } else {
            this.tabPanel.browser.open('about:blank', false);
        }
    } catch(ex) {
        log.exception(ex);
    }
}


/*
  called when the state of the session changes
*/
DebugSessionTabManager.prototype.updateUI = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].updateUI()");
    try {
        var state = ko.dbg.manager.STATE_STOPPED;
        var language = null;
        if (this.session) {
            state = this.session.getState();
            if (this.session.dbg)
                language = this.session.dbg.languageName;
        }
        this.tab.setAttribute('image',
                              ko.dbg.tabManager.getTabIcon(state, language));

        // always update our thread and call stack.  We'll only refill the callstack
        // if the current session is in break mode.  There is the potential that a thread
        // will change a value on us, but not much we can do about that without a lot
        // more overhead
        if (!this.isInteractive)
            this.buildThreadMenu();

        this.clearCallStackList();

        switch(state) {
        case ko.dbg.manager.STATE_STOPPED:
            this.tabPanel.infoText.value = "Debugging session has ended.";
            break;

        case ko.dbg.manager.STATE_STARTING:
            this.tabPanel.infoText.value = "Debugger is starting...";
            break;

        case ko.dbg.manager.STATE_RUNNING:
            this.tabPanel.infoText.value = "Debugger is running...";
            break;

        case ko.dbg.manager.STATE_INTERACTIVE:
            if (this.isInteractive) {
                if (!this.tabPanel.getElement('vbox','varViewers').getAttribute('collapsed'))
                    this.updateVariables();
                break;
            }
        case ko.dbg.manager.STATE_BREAK:
            this.buildCallStackList();

            var frame, haveFrame = false;
            if (this.stackFrames) {
                frame = this.stackFrames[0];
                if (frame)
                    haveFrame = true;
            }
            if (haveFrame)
                this.tabPanel.infoText.value =  "Break at " +
                                ko.uriparse.displayPath(frame.filename)
                                + ", line " + frame.lineno + ".";
            else
                this.tabPanel.infoText.value =  "";
            break;

        case ko.dbg.manager.STATE_STOPPING:
            this.tabPanel.infoText.value = "Debugger is stopping...";
            break;

        } // end switch(new_state)
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.clearThreadMenu = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].clearThreadMenu()");
    try {
        /* clear the thread list */
        this.threads = null;
        while (this.tabPanel.threadsMenuPopup.childNodes.length > 0) {
            this.tabPanel.threadsMenuPopup.removeChild(this.tabPanel.threadsMenuPopup.lastChild);
        }
        this.tabPanel.threadsBox.setAttribute('collapsed',true);

        this.tabPanel.threadsMenuList.value = null;



    } catch(ex) {
        log.exception(ex);
    }
}

/* reset the threads list */
DebugSessionTabManager.prototype.buildThreadMenu = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].buildThreadMenu()");
    try {
        this.clearThreadMenu();
        if (!this.session || !this.session.application)
            return;

        var tmp = new Object();
        this.session.application.getSessionList(tmp, new Object());
        this.threads = tmp.value;
        if (this.threads.length < 2)
            return;

        this.tabPanel.threadsBox.removeAttribute('collapsed');
        var item, thread;
        for (var i = 0; i < this.threads.length; ++i) {
            thread = this.threads[i];
            item = document.createElementNS(NS_XUL, 'menuitem');
            item.setAttribute('class','menuitem-iconic');
            item.setAttribute('label', thread.threadId);
            var image = ko.dbg.tabManager.getTabIcon(thread.status, thread.languageName);
            item.setAttribute('src', image);
            item.setAttribute('image', image);
            item.setAttribute('oncommand','this.value.selectSessionThread('+i+');');
            item.setAttribute('flex','1');
            item.value = this;
            this.tabPanel.threadsMenuPopup.appendChild(item);
            if (thread == this.session.dbg) {
                this.tabPanel.threadsMenuList.selectedIndex = i;
            }
        }
        if (this.tabPanel.threadsMenuList.selectedIndex < 0)
            this.tabPanel.threadsMenuList.selectedIndex = 0;
        this.currentThreadIndex = this.tabPanel.threadsMenuList.selectedIndex;
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.selectSessionThread = function(index)
{
    /* this is to allow the thread dropdown menu to release, since selecting
        a different thread can result in rebuilding the menu also */
    this._timeout = new objectTimer(this,this._selectSessionThread,[index]);
    this._timeout.startTimeout(1);
}

DebugSessionTabManager.prototype._selectSessionThread = function(index)
{
    log.debug("DebugSessionTabManager["+this.id+"].selectSessionThread("+index+")");
    try {
        if (index == this.currentThreadIndex) {
            log.debug("  already selected: "+index);
            return;
        }
        this.currentThreadIndex = index;
        this.session.application.currentSession = this.threads[index];
        if (this.session.getState() == ko.dbg.manager.STATE_RUNNING) {
            // cannot do stack for running threads
            this.clearCallStackList();
            return;
        }
        // reset the UI for the new thread selection
        this.session.updateUI();
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.clearCallStackList = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].clearCallStackList()");
    try {
        /* clear the thread list */
        this.stackFrames = null;

        while (this.tabPanel.callstackListbox.childNodes.length > 1) {
            this.tabPanel.callstackListbox.removeChild(this.tabPanel.callstackListbox.childNodes[1]);
        }
    } catch(ex) {
        log.exception(ex);
    }
}

/* reset the threads list */
DebugSessionTabManager.prototype.buildCallStackList = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].buildCallStackList()");
    try {
        this.clearCallStackList();
        var state = this.session.getState();
        if (state != ko.dbg.manager.STATE_BREAK && state != ko.dbg.manager.STATE_INTERACTIVE) {
            return;
        }
        var tmp = new Object();
        this.session.dbg.stackFramesGet(tmp, new Object());
        this.stackFrames = tmp.value;
        if (this.stackFrames.length < 1) {
            if (state == ko.dbg.manager.STATE_INTERACTIVE) {
                this.tab.setAttribute('label', this.session.dbg.languageName+' Shell');
            }
            return;
        }
        if (state == ko.dbg.manager.STATE_INTERACTIVE) {
            this.tab.setAttribute('label', this.session.dbg.languageName+' Shell');
        } else {
            this.name = ko.uriparse.baseName(this.stackFrames[0].filename);
            this.tab.setAttribute('label', 'Debug: '+this.name);
        }

        var item, frame;
        for (var i = 0; i < this.stackFrames.length; ++i) {
            frame = this.stackFrames[i];
            var frameName = ko.uriparse.baseName(frame.filename) +
                        ', line ' + frame.lineno +
                        ', in ' + frame.where;
            item = this.tabPanel.callstackListbox.appendItem(frameName, this)
            item.setAttribute('style', 'margin: 0px; border:none');
            item.setAttribute('class','listitem-iconic');
            item.value = this;
        }
        this.tabPanel.callstackListbox.selectedIndex = 0;
        this.currentStackFrameIndex = 0;
        this.doStackFrame();

        this.tabPanel.callstackListbox.selectedItem.setAttribute('currentview','true');
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.selectStackFrame = function(index)
{
    log.debug("DebugSessionTabManager["+this.id+"].selectStackFrame("+index+")");
    try {
        if (index == this.currentStackFrameIndex) {
            log.debug("  already selected: "+index);
            return;
        }
        if (this.session.getState() == ko.dbg.manager.STATE_RUNNING) {
            // cannot do stack for running threads
            return;
        }
        var old = this.tabPanel.callstackListbox.getItemAtIndex(this.currentStackFrameIndex);
        old.removeAttribute('currentview');

        this.currentStackFrameIndex = index;
        this.doStackFrame();

        this.tabPanel.callstackListbox.selectedItem.setAttribute('currentview','true');
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.doStackFrame = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].doStackFrame("+this.currentStackFrameIndex+")");
    try {
        // rebuild the variable lists
        this.updateVariables();
        this.session.showStackPosition(this.stackFrames[this.currentStackFrameIndex],
                                       this.currentStackFrameIndex);
    } catch(ex) {
        log.exception(ex);
    }
}

/* these functions provide state for variable viewers between stack
   changes and sessions in the same tab.  Currently they only store
   whether the variable is opened or not in the viewers. */
DebugSessionTabManager.prototype.getVarState = function(expression)
{
    log.debug("DebugSessionTabManager["+this.id+"].getVarState("+expression+")");
    try {
        if (typeof(this._varState[expression]) != 'undefined')
            return this._varState[expression];
    } catch(ex) {
        log.exception(ex);
    }
    return 0;
}

DebugSessionTabManager.prototype.contractVar = function(expression)
{
    log.debug("DebugSessionTabManager["+this.id+"].contractVar("+expression+")");
    try {
        this._varState[expression] = 0
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.expandVar = function(expression)
{
    log.debug("DebugSessionTabManager["+this.id+"].expandVar("+expression+")");
    try {
        this._varState[expression] = 1
    } catch(ex) {
        log.exception(ex);
    }
}

// Method template:
//DebugSessionTabManager.prototype.foobar = function()
//{
//    log.debug("DebugSessionTabManager["+this.id+"].foobar()");
//    try {
//    } catch(ex) {
//        log.exception(ex);
//    }
//}


//---- module (i.e. not tab-specific) interface routines

//XXX This is not currently being called: it doesn't do anything yet, so
//    it doesn't matter.
this.tabManager = null;

this.tabInit =
function DebugSessionTab_OnLoad()
{
    log.debug("DebugSessionTab_OnLoad()");
    try {
        ko.dbg.tabManager = new outputTabManager();
    } catch(ex) {
        log.exception(ex);
    }
}

function outputTabManager() {
}
outputTabManager.prototype = {
    tabManagers: {},

    // Get an appropriate tab icon for the current debug state and language.
    getTabIcon: function(state, language /*=null*/)
    {
        switch(state) {
        case ko.dbg.manager.STATE_STARTING:
            return "chrome://komodo/skin/images/debug_continue_hover.png";
        case ko.dbg.manager.STATE_STOPPING:
            return "chrome://komodo/skin/images/debug_stop_hover.png";
        case ko.dbg.manager.STATE_STOPPED:
            return "chrome://komodo/skin/images/debug_stop.png";
        case ko.dbg.manager.STATE_RUNNING:
            return "chrome://komodo/skin/images/debug_continue.png";
        case ko.dbg.manager.STATE_BREAK:
            return "chrome://komodo/skin/images/debug_break.png";
        case ko.dbg.manager.STATE_INTERACTIVE:
            switch(language) {
            case "Python":
                return "chrome://komodo/skin/images/ishell_python.png";
            case "Tcl":
                return "chrome://komodo/skin/images/ishell_tcl.png";
            case "Perl":
                return "chrome://komodo/skin/images/ishell_perl.png";
            default:
                return "chrome://komodo/skin/images/ishell.png";
            }
        }
        return '';
    },

    // Uncollapsed an existing debug session tab and tabpanel for use.
    //
    //  "id" is the unique id (it is a number) for the tab.
    //
    uncollapse: function(id)
    {
        log.debug("outputTabManager.uncollapse()");
        try {
            var tabId = "debugsession"+id+"-tab";
            var tabPanelId = "debugsession"+id+"-tabpanel";

            var tab = document.getElementById(tabId);
            var tabPanel = document.getElementById(tabPanelId);

            if (!tab || !tabPanel) {
                throw("Could not uncollapsed debug session tab "+id+": elements '"+
                      tabId+"' and/or '"+tabPanelId+"' do not exist.");
            }

            if (tab.hasAttribute("collapsed"))
                tab.removeAttribute("collapsed");
            if (tabPanel.hasAttribute("collapsed"))
                tabPanel.removeAttribute("collapsed");
        } catch(ex) {
            log.exception(ex);
        }
    },

    // Create a new debug session tab.
    //
    //  "id" is the unique id (it is a number) for the tab.
    //
    createTab: function(id)
    {
        log.debug("outputTabManager.createTab()");
        try {
            var tabs = document.getElementById("output_tabs");
            var panels = document.getElementById('output_tabpanels');
            var tabPanel = document.createElementNS(NS_XUL, 'DebuggerPanel');
            tabPanel.setAttribute('id','debugsession'+id+'-tabpanel');
            var tab = document.createElementNS(NS_XUL, 'tab');
            tab.setAttribute('id','debugsession'+id+'-tab');
            tab.setAttribute('tabid',id);
            tab.setAttribute('label','Output');
            tab.setAttribute('crop', 'center');
            tab.setAttribute('image', this.getTabIcon(ko.dbg.manager.STATE_STOPPED, null));
            panels.appendChild(tabPanel);
            tabs.appendChild(tab);
            tabs.selectedItem = tab;
        } catch(ex) {
            log.exception(ex);
        }
    },

    /* Give me the next tabMgr created for a particular language (in interactive
      mode) that is still active (start looking after the currently selected
      tab if the tab group has focus).*/
    getInteractiveShell: function(language)
    {
        try {
            for (var id = 0; id < MAX_DEBUG_SESSION_TABS; ++id) {
                // Skip over free tabs, tabs w/ no sessions, or non-interactive
                // tabs.
                if (typeof(this.tabManagers[id]) == "undefined" ||
                    this.tabManagers[id].session == null ||
                    ! this.tabManagers[id].isInteractive ||
                    this.tabManagers[id].session.language != language) {
                    continue;
                }
                return this.tabManagers[id];
            }
        } catch (e) {
            log.exception(e);
        }
        return null;
    },

    // Return a usable (not busy, cleared) Debug Session Tab manager.
    getTabMgr: function()
    {
        log.debug("outputTabManager.getTabMgr()");
        try {
            // Find a free tab id.
            for (var id = 0; id < MAX_DEBUG_SESSION_TABS; ++id) {
                if (typeof(this.tabManagers[id]) == "undefined") {
                    // This is a free tab id, we will dynamically create the
                    // tab and tabpanel and use that.
                    this.createTab(id);
                    break;
                } else if (this.tabManagers[id].session == null) {
                    // Manager slot is null, there is a collapsed debug session
                    // tab and tabpanel that can be used.
                    this.uncollapse(id);
                    return this.tabManagers[id];
                    break;
                } else {
                    // Manager slot is non-null (presumably a
                    // DebugSessionTabManager): leave it alone.
                }
            }
            if (id >= MAX_DEBUG_SESSION_TABS) {
                ko.dialogs.alert("Too many debug session tabs are in use. You must "+
                             "close one or more debug session tabs and restart "+
                             "this debug session.");
                return null;
            }

            // Create a manager for this tab and return it.
            var manager = new DebugSessionTabManager(id);
            this.tabManagers[id] = manager;
            return manager;
        } catch(ex) {
            log.exception(ex);
        }
        return null;
    },

    /* we just want to create a new tab, but not attach a
       manager to it */
    showNewTab: function(tabMgr)
    {
        log.debug("outputTabManager.getNewTab()");
        try {
            if (typeof(tabMgr) == 'undefined' || !tabMgr)
                tabMgr = this.getTabMgr();
            if (!tabMgr) {
                return;
            }
            tabMgr.configure();
            tabMgr.show();
            ko.dbg.manager.setCurrentSession(null);
            this.releaseTabMgr(tabMgr);
        } catch(ex) {
            log.exception(ex);
        }
    },

    releaseTabMgr: function(tabMgr)
    {
        // nothing to do now
    },

    getCurrentTab: function()
    {
        log.debug("outputTabManager.getCurrentTab()");
        try {
            var tabs = document.getElementById("output_tabs");
            var tab = tabs.selectedItem;
            var id = tab.getAttribute('tabid');
            return this.tabManagers[id];
        } catch(ex) {
            log.exception(ex);
        }
        return null;
    },
    
    getNextTab: function(tab)
    {
        log.debug("outputTabManager.getNextTab()");
        try {
            if (typeof(tab) != 'undefined' && tab) {
                var newTab = tab.nextSibling;
                var selTab = null;
                while (newTab &&
                       newTab.getAttribute('collapsed') == 'true')
                        newTab = newTab.nextSibling;
                if (!newTab || newTab.getAttribute('collapsed') == 'true') {
                    newTab = tab.previousSibling;
                    while (newTab &&
                           newTab.getAttribute('collapsed') == 'true')
                            newTab = newTab.previousSibling;
                }
                if (newTab &&
                    newTab.getAttribute('collapsed') != 'true')
                    return newTab;
            }
        } catch(ex) {
            log.exception(ex);
        }
        return document.getElementById("output_tabs").firstChild;
    },

    closeCurrent: function()
    {
        log.debug("outputTabManager.closeCurrent()");
        try {
            var tabs = document.getElementById("output_tabs");
            var panels = document.getElementById('output_tabpanels');
            var tab = tabs.selectedItem;
            var id = tab.getAttribute('tabid');
            // if this is not a managed tab (eg. command output)
            // then it will not have a tabid attribute
            if (!id) {
                log.debug(" attempt to close tab without id");
                return;
            }
            var tabMgr = this.tabManagers[id];
            if (!tabMgr) return;

            // XXX we could call a canclose function here, and validate
            // that the user wants to shut down the debug session,
            // for now, the easy way out.
            if (!tabMgr.canClose()) {
                return;
            }

            delete this.tabManagers[id];
            tabMgr.forceClose();

            var tabId = "debugsession"+id+"-tab";
            var tabPanelId = "debugsession"+id+"-tabpanel";
            tab = document.getElementById(tabId);
            var nextTab = this.getNextTab(tab);
            tabs.removeChild(tab);
            var tabPanel = document.getElementById(tabPanelId);
            panels.removeChild(tabPanel);
            tabs.selectedItem = nextTab;
        } catch(ex) {
            log.exception(ex);
        }
    },

    tabSwitch: function(tab)
    {
        log.debug("outputTabManager.tabSwitch("+tab.id+")");
        try {
            var id = tab.getAttribute('tabid');
            // if this is not a managed tab (eg. command output)
            // then it will not have a tabid attribute
            if (!id || typeof(this.tabManagers[id]) == 'undefined') return;
            ko.dbg.manager.setCurrentSession(this.tabManagers[id].session);
        } catch(ex) {
            log.exception(ex);
        }
    }

}

}).apply(ko.dbg);
